import argparse
import json
import os
import re
import shutil
import subprocess
import sys
import rpm
import time
import tempfile
import hashlib
from ruamel.yaml import YAML
from typing import Optional
import requests

preprocess_script = "/usr/libexec/rpm-check/prepkg.sh"


class auto_pkg_ci:
    def __init__(self, token: str):

        from ocs_ci_api.models.git_api.gitee_api import GiteeAPI as git

        self.git_api = git(token)
        self.src_dir = "src_rpm"
    
    def py_fork_repo(self, package_name, group):
        print(f"Starting to fork repo {group}/{package_name} ....")
        try:
            self.git_api.fork_repo(package_name, group)
        except Exception as e:
            print(f"Errors when fork repo: {str(e)}")
    
    def py_delete_repo(self, package_name, group):
        print(f"Starting to delete repo {group}/{package_name} ....")
        try:
            self.git_api.delete_repo(package_name, group)
        except Exception as e:
            print(f"Errors when delete repo: {str(e)}")
    
    def py_create_pr(self, temp_dir2, group, package_name, username):
        print(f"Starting to create pr to {group}/{package_name} with {username}....")
        title = f"【自动打包】{package_name}"
        pr_info = self.git_api.create_pr(package_name, group, title, f"{username}:master", "master")
    
    def create_gitee_repo(self, repo_name):
        group_name = "opencloudos-stream"
        print(f"Starting creating repo {repo_name}..")
        try:
            self.git_api.create_repo(repo_name, group_name)
        except Exception as e:
            print(f"创建仓库时发生错误: {str(e)}")
    
    def update_repo(self, temp_dir: str, repo_name: str):
        print(f"Starting to update git repo....")
        repo_dir = f"{temp_dir}/{repo_name}"
        srpm_path = f"{temp_dir}/{self.src_dir}"
        if os.path.exists(repo_dir):
            # Remove all files in the repo
            for file in os.listdir(repo_dir):
                if file != '.git':
                    os.remove(os.path.join(repo_dir, file))

            for file in os.listdir(srpm_path):
                shutil.move(os.path.join(srpm_path, file), os.path.join(repo_dir, file))

            # Add, commit and push changes
            subprocess.run("git add .", shell=True, cwd=repo_dir, check=True)
            subprocess.run('git commit -m "Initial build"', shell=True, cwd=repo_dir, check=True)
            subprocess.run("git push", shell=True, cwd=repo_dir, check=True)
    
    def add_pkg_to_rpm_info(self, pkg_name, user_name, temp_dir: str):
        # add pkg to packages.yaml, this step will create repo
        self.py_delete_repo("rpm_info", user_name)
        self.py_fork_repo("rpm_info", "OpenCloudOS")
        self.clone_repo(temp_dir, f"git@gitee.com:{user_name}/rpm_info.git", "rpm_info")
        self.update_packages_yaml(pkg_name, temp_dir)
        self.py_create_pr(temp_dir, "OpenCloudOS", "rpm_info", user_name)
    
    def update_packages_yaml(self, pkg_name: str, temp_dir: str):
        # 定义文件路径
        file_path = os.path.join(temp_dir, 'rpm_info', 'packages.yaml')

        yaml = YAML()

        with open(file_path, 'r') as file:
            data = yaml.load(file)

        for item in data:
            if 'sig' in item and item['sig'] == 'default':
                item['packages'].append(pkg_name)
                break

        with open(file_path, 'w') as file:
            yaml.dump(data, file)

        # 定义Git命令
        git_add = ['git', 'add', 'packages.yaml']
        git_commit = ['git', 'commit', '-m', f'add {pkg_name}']
        git_push = ['git', 'push']

        # 执行Git命令
        try:
            subprocess.check_call(git_add, cwd=os.path.join(temp_dir, 'rpm_info'))
            subprocess.check_call(git_commit, cwd=os.path.join(temp_dir, 'rpm_info'))
            subprocess.check_call(git_push, cwd=os.path.join(temp_dir, 'rpm_info'))
        except subprocess.CalledProcessError as e:
            print(f"执行Git命令时发生错误: {str(e)}")
    
    def clone_repo(self, temp_dir: str, git_repo_url: str, git_repo_name: str):
        print(f"Starting to clone git repo....")
        cmd = f"git clone {git_repo_url} {temp_dir}/{git_repo_name}/"
        subprocess.run(cmd, shell=True, check=True)
    
    def pkg_repo_exist(self, url: str) -> bool:
        try:
            response = requests.get(url)
            if response.status_code == 404:
                print(f"{url} not exists: {response.status_code}")
                return False
            else:
                return True
        except requests.RequestException as e:
            print(f"An error occurred when request repo url: {e}.")
    
    def check_url_until_success(self, url, wait_time=5):
        """
        Continuously check the URL until it returns a status code other than 404.

        :param url: The URL to check.
        :param wait_time: Time to wait between retries in seconds.
        """

        while True:
            try:
                response = requests.get(url)
                if response.status_code != 404:
                    print(f"Success! Status code: {response.status_code}")
                    time.sleep(5)
                    break
                else:
                    print(f"Received 404. Retrying in {wait_time} seconds...")
            except requests.RequestException as e:
                print(f"An error occurred: {e}. Retrying in {wait_time} seconds...")
            
            time.sleep(wait_time)

class PyPkg:
    def __init__(self, pypi_name: str):
        self.pypi_name = pypi_name
        self.version = None
        self.license = None
        self.user_name = self.get_git_username()
        self.src_dir = "src_rpm"

        url = f"https://pypi.org/pypi/{self.pypi_name}/json"
        response = requests.get(url)

        if response.status_code == 200:
            data = json.loads(response.text)
            self.version = data.get("info", {}).get("version")
            self.license = self.extract_license(data.get("info", {}).get("classifiers"))
        else:
            print(f"Error: Unable to fetch metadata for {self.pypi_name}")

    def extract_license(self, classifiers):
        for classifier in classifiers:
            if classifier.startswith("License ::"):
                # 提取许可证信息
                license_info = classifier.split("::")[-1].strip()
                # 确保许可证信息只包含有效字符
                if re.match(r'^[\w\s\.\:\-]+$', license_info):
                    return license_info
                else:
                    raise ValueError(f"Invalid license key: {license_info}")
        return None

    def get_git_username(self) -> Optional[str]:
        try:
            username = subprocess.check_output(['git', 'config', '--get', 'user.name'])
            return username.decode('utf-8').strip()
        except subprocess.CalledProcessError:
            print("Error: Could not get Git username")
            return None

    def build_srpm(self) -> Optional[str]:
        if not self.version:
            print(f"Error: Unable to build SRPM for {self.pypi_name} due to missing version")
            return None

        dist_name = rpm.expandMacro('%dist')

        cmd = f"pyp2rpm {self.pypi_name} --srpm -v {self.version}"
        subprocess.run(cmd, shell=True, check=True)

        srpm_path = f"/root/rpmbuild/SRPMS/python-{self.pypi_name}-{self.version}-1{dist_name}.src.rpm"
        if os.path.exists(srpm_path):
            return srpm_path
        else:
            print(f"Error: Unable to find the SRPM file at {srpm_path}")
            return None
    
    def build_spec(self) -> Optional[str]:
        if not self.version:
            print(f"Error: Unable to build SPEC for {self.pypi_name} due to missing version")
            return None

        cmd = f"pyp2spec {self.pypi_name} -v {self.version} -l '{self.license}'"
        subprocess.run(cmd, shell=True, check=True)

        spec_path = f"./python-{self.pypi_name}.spec"
        if os.path.exists(spec_path):
            return spec_path
        else:
            print(f"Error: Unable to find the SPEC file created by {self.pypi_name}")
            return None

    def download_deps(self, srpm_path: str, temp_dir: str) -> bool:
        cmd = f"dnf builddep --downloadonly {srpm_path} --downloaddir {temp_dir} -y"
        result = subprocess.run(cmd, shell=True)

        if result.returncode == 0:
            return True
        else:
            print(f"Error: Dependencies not satisfied for {srpm_path}")
            return False

    def remove_comment(self, file_name: str) -> None:
        print(f"SPEC file with comments removed: {file_name}")

        with open(file_name, "r") as f:
            content = f.readlines()

        content = [line for line in content if not line.startswith("#")]

        with open(file_name, "w") as f:
            f.writelines(content)

    def preprocess_spec(self, file_name: str):
        print(f"SPEC file preprocessed: {file_name}")
        subprocess.run([preprocess_script, file_name], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    
    def classify_spec(self, file_name: str) -> None:
        self.remove_comment(file_name)
        self.preprocess_spec(file_name)

    def extract_spec_file(self, srpm_path: str, tmpdir: str) -> Optional[str]:
        os.makedirs(f"{tmpdir}/{self.src_dir}")
        subprocess.run(f"rpm2cpio {srpm_path} | cpio -idmv -D {tmpdir}/{self.src_dir}", shell=True, check=True)
        spec_file = f"{tmpdir}/{self.src_dir}/python-{self.pypi_name}.spec"

        if os.path.exists(spec_file):
            self.classify_spec(spec_file)
            return spec_file
        else:
            print(f"Error: Unable to find the SPEC file at {spec_file}")
            return None
    
    def calculate_sha512(filename):
        sha512_hash = hashlib.sha512()

        with open(filename,"rb") as f:
            for byte_block in iter(lambda: f.read(4096),b""):
                sha512_hash.update(byte_block)

        return sha512_hash.hexdigest()

    def write_to_sources(filename, sha512sum):
        with open('sources', 'w') as f:
            f.write(f'SHA512 ({filename}) = {sha512sum}\n')
    
    def get_source(self, spec_file: str, dest_dir: str):
        url = self.get_source_url(spec_file)
        if url:
            print(f"Source URL: {url}")
            self.download_file(url, dest_dir)

    def get_source_url(self, spec_file):
        try:
            # 使用 rpmspec -P 命令解析 SPEC 文件
            result = subprocess.run(['rpmspec', '-P', spec_file], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            
            if result.returncode != 0:
                print(f"Error running rpmspec: {result.stderr}")
                return None
            
            # 查找包含 Source: 的行
            for line in result.stdout.splitlines():
                if line.startswith("Source:"):
                    # 使用正则表达式提取 URL
                    match = re.search(r'Source:\s*(\S+)', line)
                    if match:
                        return match.group(1)
            
            print("No Source URL found in the spec file.")
            return None
        
        except Exception as e:
            print(f"An error occurred when parse SPEC: {e}")
            return None

    def download_file(self, url, dest_dir):
        try:
            # 确保目标目录存在
            os.makedirs(dest_dir, exist_ok=True)
            
            # 使用 wget 下载文件
            result = subprocess.run(['wget', '-P', dest_dir, url], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
            
            if result.returncode != 0:
                print(f"Error downloading file: {result.stderr}")
            else:
                print(f"File downloaded successfully to {dest_dir}")
        
        except Exception as e:
            print(f"An error occurred when download src: {e}")


def move_spec_to_temp_dir(src_file, dest_dir):
    try:
        # 确保目标目录存在，如果不存在则创建
        os.makedirs(dest_dir, exist_ok=True)
        
        # 获取文件名
        file_name = os.path.basename(src_file)
        
        # 构建目标文件路径
        dest_file = os.path.join(dest_dir, file_name)
        
        # 移动文件
        shutil.move(src_file, dest_file)
        
        print(f"File '{src_file}' has been moved to '{dest_file}'")
    
    except Exception as e:
        print(f"An error occurred when move file: {e}")


def main():
    # 创建 ArgumentParser 对象
    parser = argparse.ArgumentParser()

    # 添加 pypi_name 参数
    parser.add_argument("pypi_name", help="The name of the PyPI package")

    # 创建一个互斥组，确保 --pr 和 --token 一起使用
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument("--pr", action='store_true', help="A boolean flag to indicate something")

    # 添加 --token 参数，作为可选参数
    parser.add_argument("--token", help="The token for the Git API", required='--pr' in sys.argv)

    parser.add_argument("--skip-req", action='store_true', help="skip the requirement failed")

    parser.add_argument("--compat", action='store_true', help="use new approach to generate spec")

    # 解析命令行参数
    args = parser.parse_args()

    pkg = PyPkg(args.pypi_name)

    pkg_git_url = f"git@gitee.com:opencloudos-stream/python-{pkg.pypi_name}.git"
    pkg_repo_url = f"https://gitee.com/opencloudos-stream/python-{pkg.pypi_name}"

    if pkg.version:
        print(f"Version of {args.pypi_name}: {pkg.version}")

        srpm_path = None
        spec_path = None

        if not args.compat:
            spec_path = pkg.build_spec()
        else:
            try:
                srpm_path = pkg.build_srpm()
            except Exception as e:
                print(f"Error happened when using pyp2rpm: {e}")
                spec_path = pkg.build_spec()

        if srpm_path:
            print(f"SRPM file created: {srpm_path}")

            with tempfile.TemporaryDirectory() as temp_dir:
                if not pkg.download_deps(srpm_path, temp_dir):
                    if not args.skip_req:
                        sys.exit(1)

            with tempfile.TemporaryDirectory() as temp_dir2:
                spec_file = pkg.extract_spec_file(srpm_path, temp_dir2)

                if not args.pr:
                    return
                
                from ocs_ci_api.models.git_api.gitee_api import GiteeAPI as git
                
                py_ci = auto_pkg_ci(args.token)

                if spec_file and os.path.exists(spec_file):

                    if not py_ci.pkg_repo_exist(pkg_repo_url):
                        py_ci.add_pkg_to_rpm_info(f"python-{pkg.pypi_name}", pkg.user_name, temp_dir2)

                        # wait for CI to add repo
                        py_ci.check_url_until_success(pkg_repo_url)
                    
                    py_ci.py_delete_repo(f"python-{pkg.pypi_name}", pkg.user_name)
                    py_ci.py_fork_repo(f"python-{pkg.pypi_name}", "opencloudos-stream")
                    py_ci.clone_repo(temp_dir2, f"git@gitee.com:{pkg.user_name}/python-{pkg.pypi_name}.git", f"python-{pkg.pypi_name}")
                    py_ci.update_repo(temp_dir2, f"python-{pkg.pypi_name}")
                    py_ci.py_create_pr(temp_dir2, "opencloudos-stream", f"python-{pkg.pypi_name}", pkg.user_name)
                else:
                    print(f"SPEC file {spec_file} not exist!")
        elif spec_path:
            print(f"SPEC file created: {spec_path}")

            with tempfile.TemporaryDirectory() as temp_dir2:

                pkg.classify_spec(spec_path)

                if not args.pr:
                    return
                
                py_ci = auto_pkg_ci(args.token)

                pkg.get_source(spec_path, os.path.join(temp_dir2, pkg.src_dir))

                move_spec_to_temp_dir(spec_path, os.path.join(temp_dir2, pkg.src_dir))

                if not py_ci.pkg_repo_exist(pkg_repo_url):
                    py_ci.add_pkg_to_rpm_info(f"python-{pkg.pypi_name}", pkg.user_name, temp_dir2)

                    # wait for CI to add repo
                    py_ci.check_url_until_success(pkg_repo_url)
                
                py_ci.py_delete_repo(f"python-{pkg.pypi_name}", pkg.user_name)
                py_ci.py_fork_repo(f"python-{pkg.pypi_name}", "opencloudos-stream")
                py_ci.clone_repo(temp_dir2, f"git@gitee.com:{pkg.user_name}/python-{pkg.pypi_name}.git", f"python-{pkg.pypi_name}")
                py_ci.update_repo(temp_dir2, f"python-{pkg.pypi_name}")
                py_ci.py_create_pr(temp_dir2, "opencloudos-stream", f"python-{pkg.pypi_name}", pkg.user_name)

if __name__ == "__main__":
    main()
