import importlib.metadata
from importlib.metadata import distribution
import argparse
import subprocess
import json
import os
import sys
import re

# 获取内置模块列表
builtin_modules = set(sys.builtin_module_names)


def uncanonicalize_name(name: str):
    # reverse packaging.utils.canonicalize_name
    _canonicalize_regex = re.compile(r"[-_.]+")
    value = _canonicalize_regex.sub("_", name).lower()
    return value

def get_standard_library_modules():
    lib_path = os.path.dirname(os.__file__)
    modules = []

    def add_module(root, file):
        module_path = os.path.relpath(os.path.join(root, file), lib_path)
        module_name = os.path.splitext(module_path.replace(os.path.sep, '.'))[0]
        if module_name.endswith('.__init__'):
            module_name = module_name[:-9]
        modules.append(module_name)

    for root, dirs, files in os.walk(lib_path):
        if 'site-packages' in dirs:
            dirs.remove('site-packages')
        if root == lib_path:
            # 获取第一层的所有 .py 文件名
            for file in files:
                if file.endswith('.py'):
                    add_module(root, file)

        # 处理带有 __init__.py 文件的目录链
        if '__init__.py' in files:
            add_module(root, '__init__.py')

    return modules

# 添加一些常见的标准库模块
builtin_modules.update(get_standard_library_modules())


def find_record_file(package_name):
    """找到指定包的 RECORD 文件路径"""
    dist = distribution(package_name)
    dist_info_dir = os.path.join(dist.locate_file(""), f"{uncanonicalize_name(dist.metadata['name'])}-{dist.version}.dist-info")
    record_file = os.path.join(dist_info_dir, "RECORD")
    if os.path.exists(record_file):
        return record_file
    else:
        print(f"RECORD 文件未找到: {record_file}")
        return None

def parse_record_file(record_file):
    """解析 RECORD 文件，提取顶级模块名"""
    top_level_modules = set()
    with open(record_file, "r") as f:
        for line in f:
            file_path = line.split(",")[0]
            if file_path.endswith(".py"):
                top_level_module = file_path.split("/")[0]
                # /usr/local/lib/python3.11/site-packages/six.py
                if top_level_module.endswith(".py"):
                    top_level_module = top_level_module[:-3]
                top_level_modules.add(top_level_module)
    return top_level_modules

def get_top_level_modules(package_name) -> set():
    record_file = find_record_file(package_name)
    if record_file:
        top_level_modules = parse_record_file(record_file)
        # print(f"包 {package_name} 的顶级模块:")
        #for module in sorted(top_level_modules):
        #    print(module)
        return top_level_modules
    else:
        return set()


def parse_imports(file_path):
    with open(file_path, 'r', errors='ignore') as file:
        content = file.read()

    # 移除所有单行注释
    content = re.sub(r'#.*', '', content)

    # 移除所有多行注释
    content = re.sub(r'""".*?"""', '', content, flags=re.DOTALL)

    # 匹配 import 和 from import 语句
    import_re = re.compile(r'^\s*(?:from\s+([.\w]+)(?:\s+import\s+[\w, ()]+)|import\s+([\w, ()]+))', re.MULTILINE)
    matches = import_re.findall(content)

    imports = []
    for match in matches:
        # match 是一个元组，其中一个元素是空字符串，另一个元素是模块名
        module_names = match[0] if match[0] else match[1]
        # 如果模块名以'.'开头，说明是相对导入，我们忽略它
        if not module_names.startswith('.'):
            module_names = module_names.split(',')
            for module_name in module_names:
                # 处理别名导入的情况
                module_name = module_name.strip().split(' as ')[0].split('.')[0]
                if module_name not in builtin_modules and not module_name.startswith('_'):
                    imports.append(module_name)

    return imports

def get_package_imports():
    package_imports = {}


    dists = importlib.metadata.distributions()
    for dist in dists:
        package_name = dist.metadata['Name']

        if not package_name:
            continue

        package_imports[package_name] = {}

        try:
            import pkg_resources
            package_path = pkg_resources.resource_filename(package_name, '')

            for root, dirs, files in os.walk(package_path):
                for file in files:
                    if file.endswith('.py'):
                        file_path = os.path.join(root, file)
                        imports = parse_imports(file_path)
                        # 去重并筛除当前包名
                        imports = list(set(imports))
                        if package_name in imports:
                            imports.remove(package_name)
                        package_imports[package_name][file_path] = imports

            continue
            
        except:
            #print(f"{package_name} is not found by pkg_resources")
            pass

        
        if not distribution(package_name).files:
            print(f"no file found in {package_name}.")
            continue

        for file_path in distribution(package_name).files:
            absolute_file_path = str(distribution(package_name).locate_file(file_path))

            if not os.path.exists(absolute_file_path):
                #print(f"{absolute_file_path} of {package_name} not found.")
                continue

            if absolute_file_path.endswith('.py'):
                #print(absolute_file_path)
                imports = parse_imports(absolute_file_path)
                # 去重并筛除当前包名
                imports = list(set(imports))
                if package_name in imports:
                    imports.remove(package_name)
                package_imports[package_name][absolute_file_path] = imports

    return package_imports

def check_py_src(package_list: set()):
    # 获取所有包的import信息
    package_imports = get_package_imports()

    #print(package_imports)
    
    check_packages = set()

    for pkg in package_list:
        ret = get_pipdeptree_json(pkg)
        for package in json.loads(ret):
            package_name = package['package']['package_name']

            if package_name in package_imports:
                dependencies = set()
                for dep in package['dependencies']:
                    dependencies.update(get_top_level_modules(dep['package_name']))            

                for file_path, imports in package_imports[package_name].items():
                    for import_name in imports:
                        if import_name not in dependencies:
                            print(f'In package {package_name}, file {file_path} imports {import_name} which is not in pipdeptree dependencies.')
                        else:
                            #print(f'In package {package_name}, file {file_path} imports {import_name} is found in pipdeptree.')
                            pass
                
                for dep in dependencies:
                    got = False
                    for file_path, imports in package_imports[package_name].items():
                        if dep in imports:
                            got = True
                    
                    if not got:
                        print(f'In package {package_name}, dep {dep} is not found in sources.')


def get_pipdeptree_json(package_name: str):
    ret = subprocess.run(['pipdeptree', '-p', package_name, '-j'], stderr=subprocess.DEVNULL, stdout=subprocess.PIPE).stdout.decode()
    return ret

def get_all_pipdeptree_json():
    ret = subprocess.run(['pipdeptree', '-j'], stderr=subprocess.DEVNULL, stdout=subprocess.PIPE).stdout.decode()
    return ret

def check_redundant_package(exclude_package: [str]) -> set():

    exclude_list = []
    exclude_pkg = set()

    all_pkg = set()

    ret = get_all_pipdeptree_json()
    for package in json.loads(ret):
        package_name = package['package']['package_name']
        all_pkg.add(package_name)

    for ex_package in exclude_package:
        ret = get_pipdeptree_json(ex_package)
        for package in json.loads(ret):
            package_name = package['package']['package_name']
            exclude_pkg.add(package_name)
    
    print("Following packages are redundant:")
    print(all_pkg - exclude_pkg)

    return exclude_pkg

def main():

    # 创建一个解析器
    parser = argparse.ArgumentParser(description='This is a tool to optimize the python dependencies')

    # 添加一个参数
    parser.add_argument('--package', '-p', type=str, help='package to exclude')

    # 解析参数
    args = parser.parse_args()

    if args.package:
        package_list = args.package.split(',')

    pkgs_to_check_src = set()
    pkgs_to_check_src = check_redundant_package(package_list)

    check_py_src(pkgs_to_check_src)

if __name__ == '__main__':
    main()