import os
import json
import re

def parse_summary_file(file_path):
    with open(file_path, "r") as f:
        lines = f.readlines()

    libraries = []
    current_library = {}

    for line in lines:
        if "Binary compatibility:" in line:
            current_library["binary_compatibility"] = line.strip().split()[2]
        elif "Source compatibility:" in line:
            current_library["source_compatibility"] = line.strip().split()[2]
        elif "Total binary compatibility problems:" in line:
            current_library["binary_compatibility_problems"] = int(line.split()[4].strip(','))
            current_library["binary_compatibility_warnings"] = int(line.split()[6])
        elif "Total source compatibility problems:" in line:
            current_library["source_compatibility_problems"] = int(line.strip().split()[4].strip(','))
            current_library["source_compatibility_warnings"] = int(line.strip().split()[6])
        elif "Report:" in line:
            current_library["library_name"] = line.strip().split("/")[1]
        elif "soname:" in line:
            libraries.append(current_library)
            current_library = {}

    return libraries

def parse_confdiff_file(file_path):
    with open(file_path, "r") as f:
        lines = f.readlines()

    packages = []
    current_package_data = {"package_name": [], "config_file": [], "options_changed": [], "options_removed": [], "options_added": [], "config_file_diff": [], "section_removed": [], "section_added": []}
    current_config_file = ""
    current_package = ""

    for line in lines:
        if line.startswith("========================================"):
            if current_package_data["config_file"] or current_package_data["options_changed"] or current_package_data["options_removed"] or current_package_data["options_added"] or current_package_data["config_file_diff"] or current_package_data["section_removed"] or current_package_data["section_added"]:
                packages.append(current_package_data)
                current_package_data = {"package_name": [], "config_file": [], "options_changed": [], "options_removed": [], "options_added": [], "config_file_diff": [], "section_removed": [], "section_added": []}
            current_package = line.strip().split()[1]
            current_package_data["package_name"] = current_package
        elif "Option" in line and "changed" in line:
            option_name = line.strip().split("'")[1]
            current_package_data["options_changed"].append({"option_name": option_name})
        elif "Option" in line and "removed" in line:
            option_name = line.strip().split("'")[1]
            current_package_data["options_removed"].append({"option_name": option_name})
        elif "Option" in line and "added" in line:
            option_name = line.strip().split("'")[1]
            current_package_data["options_added"].append({"option_name": option_name})
        elif "Section" in line and "removed" in line:
            section_name = line.strip().split("'")[1]
            current_package_data["section_removed"].append({"section_name": section_name})
        elif "Section" in line and "added" in line:
            section_name = line.strip().split("'")[1]
            current_package_data["section_added"].append({"section_name": section_name})
        elif "Configuration file differences for" in line:
            if current_package_data["config_file"] or current_package_data["options_changed"] or current_package_data["options_removed"] or current_package_data["options_added"] or current_package_data["config_file_diff"] or current_package_data["section_removed"] or current_package_data["section_added"]:
                packages.append(current_package_data)
                current_package_data = {"package_name": current_package_data["package_name"], "config_file": [], "options_changed": [], "options_removed": [], "options_added": [], "config_file_diff": [], "section_removed": [], "section_added": []}
            current_config_file = line.strip().split()[-1][:-1]
            current_package_data["config_file"] = current_config_file
        elif "---" in line or "+++" in line or "@@" in line:
            continue
        elif current_config_file and not line.startswith("========================================"):
            if line.startswith("+") or line.startswith("-") or "is missing in the" in line or "has different values:" in line:
                current_package_data["config_file_diff"].append(line.strip())

    if current_package_data["config_file"] or current_package_data["options_changed"] or current_package_data["options_removed"] or current_package_data["options_added"] or current_package_data["config_file_diff"] or current_package_data["section_removed"] or current_package_data["section_added"]:
        packages.append(current_package_data)

    return packages


def parse_rpmdiff_file(file_path):
    with open(file_path, "r") as f:
        lines = f.readlines()

    packages = []
    current_package_data = {"added": {}, "removed": {}}
    current_package_old = ""
    current_package_new = ""

    for line in lines:
        if line.startswith("=========="):
            if any(current_package_data["added"].values()) or any(current_package_data["removed"].values()):
                packages.append(current_package_data)
                current_package_data = {"added": {}, "removed": {}}
            current_package_old = os.path.basename(line.strip().split()[6])
            current_package_new = os.path.basename(line.strip().split()[9])
            current_package_data["package_old"] = current_package_old
            current_package_data["package_new"] = current_package_new 
        elif any(capability in line for capability in ["REQUIRES", "PROVIDES", "CONFLICTS", "OBSOLETES", "RECOMMENDS", "SUPPLEMENTS", "SUGGESTS"]):
            action, capability = line.strip().split()[:2]
            value = line.strip().split(capability)[1].strip()
            if action == "added":
                if capability not in current_package_data["added"]:
                    current_package_data["added"][capability] = []
                current_package_data["added"][capability].append(value)
            elif action == "removed":
                if capability not in current_package_data["removed"]:
                    current_package_data["removed"][capability] = []
                current_package_data["removed"][capability].append(value)

    if any(current_package_data["added"].values()) or any(current_package_data["removed"].values()):
        packages.append(current_package_data)

    return packages

def parse_abidiff_file(file_path):
    with open(file_path, "r") as file:
        log_lines = file.readlines()

    problems_list = []

    library_name_pattern = r"Test library: (\w+\.\w+)"
    library_name = "Unknown"
    problems = {
        "Library Name": library_name,
        "abi_diff": {"Medium": [], "High": [], "Low": []},
        "api_diff": {"Medium": [], "High": [], "Low": []}
    }

    processing_type = None
    current_severity = None
    new_lib = False

    for line in log_lines:
        library_name_match = re.search(library_name_pattern, line)
        if library_name_match:
            library_name = library_name_match.group(1)
            problems["Library Name"] = library_name

        if re.search(r"Processing (binary|source) report", line):

            if processing_type:
                problems_list.append(problems)
                new_lib = True

            processing_type_match = re.search(r"Processing (binary|source) report", line)
            processing_type = "abi_diff" if processing_type_match.group(1) == "binary" else "api_diff"

            if new_lib and processing_type == "api_diff":
                problems = {
                    "Library Name": "",
                    "abi_diff": {"Medium": [], "High": [], "Low": []},
                    "api_diff": {"Medium": [], "High": [], "Low": []}
                }
                new_lib = False

        severity_match = re.search(r"Problems with severity '(\w+)':", line)
        if severity_match:
            current_severity = severity_match.group(1)

        if processing_type and current_severity and current_severity != "Safe":
            problem_match = re.search(r"^\s{2}([\w\s]+):([\w\s.,'-]+)", line)
            if problem_match:
                key, value = problem_match.group(1).strip(), problem_match.group(2).strip()
                problems[processing_type][current_severity].append({key: value})

    problems_list.append(problems)
    return problems_list

# Find all log files in the current directory
summary_files = [f for f in os.listdir() if f.endswith("_summary.log")]
confdiff_files = [f for f in os.listdir() if f.endswith("_confdiff.log")]
abilitydiff_files = [f for f in os.listdir() if f.endswith("_rpmdiff.log")]
abidiff_files = [f for f in os.listdir() if f.endswith("_abidiff.log")]

# Group log files by package name
package_logs = {}
for file in summary_files + confdiff_files + abilitydiff_files + abidiff_files:
    pkg_name = file.split("_")[0]
    if pkg_name not in package_logs:
        package_logs[pkg_name] = []
    package_logs[pkg_name].append(file)

#print(package_logs)

# Parse log files and store the results in a list
results = []
for pkg_name, log_files in package_logs.items():
    pkg_data = {"pkg_name": pkg_name}
    for log_file in log_files:
        if "summary.log" in log_file:
            pkg_data["compatibility"] = parse_summary_file(log_file)
        elif "confdiff.log" in log_file:
            pkg_data["config_file_changes"] = parse_confdiff_file(log_file)
        elif "rpmdiff.log" in log_file:
            pkg_data["requires_changes"] = parse_rpmdiff_file(log_file)
        elif "abidiff.log" in log_file:
            pkg_data["detail_compatibility"] = parse_abidiff_file(log_file)
    results.append(pkg_data)

# Convert the results list to a JSON string and write it to a file
json_string = json.dumps(results, indent=2)
with open("output.json", "w") as f:
    f.write(json_string)